# -*- coding: utf-8 -*-
#
# Copyright (C) 2007 Edgewall Software
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://posterity.edgewall.org/wiki/License.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://posterity.edgewall.org/log/.

from datetime import datetime, timedelta
import email
import poplib
import socket
import time

from elixir import metadata, objectstore
from sqlalchemy import *

from posterity.admin import AdminCommand
from posterity.admin.inject import Injector
from posterity.model import Pop3Account, Pop3Log


class POP3Fetcher(Injector):
    """Fetch messages from external POP3 servers
    
    """
    def __init__(self, instance, once=False, interval=300):
        """Initialize a POP3Fetcher object
        
        :param instance: The Posterity instance to work on
        :param once: Check for messages once, or loop forever
        :param interval: Delay in seconds between server connections
        """
        super(POP3Fetcher, self).__init__(instance)
        self.once = once
        self.interval = interval

    def run(self):
        while True:
            t0 = time.time()
            t = datetime.utcnow() - timedelta(seconds=self.interval)
            paccounts = Pop3Account.select(and_(Pop3Account.c.enabled,
                                                or_(Pop3Account.c.last_check==None,
                                                    Pop3Account.c.last_check<t)))
            for paccount in paccounts:
                try:
                    func = lambda: self.fetch_messages(paccount)
                    self.instance.run_in_transaction(func)
                except socket.error, e:
                    paccount.error_msg = e.args[1]
                except poplib.error_proto, e:
                    paccount.error_msg = e.args[0]
                except Exception, e:
                    self.log.exception(e)
                    paccount.error_msg = 'Unknown error'
                else:
                    paccount.error_msg = None
                if paccount.error_msg:
                    self.log.warning('Operation failed: ' + paccount.error_msg)
            if self.once:
                break
            delay = self.interval - (time.time() - t0)
            if delay > 0:
                time.sleep(delay)

    def fetch_messages(self, paccount):
        """Connect to a POP3 account and download all new messages

        This function automatically keeps track of already downloaded messages.
        After all new messages have been downloaded and injected this function
        checks the `Pop3Log` for any old messages that should be deleted.
        
        :param paccount: The `Pop3Account` to connect to
        """
        now = datetime.utcnow()
        paccount.last_check = now
        self.log.debug('Connecting to POP3 server ' + paccount.hostname)
        if paccount.hostname == 'fake.example.com':
            from posterity.tests.fixtures import MockPOP3
            server = MockPOP3(paccount.hostname)
        elif paccount.use_ssl:
            server = poplib.POP3_SSL(paccount.hostname)
        else:
            server = poplib.POP3(paccount.hostname)
        server.user(paccount.username)
        server.pass_(paccount.password)
        downloaded = dict([(log.muid, log.date) for log in paccount.log])
        respone, msgs, octets = server.uidl()
        self.log.debug('POP3 mailbox contains %d messages' % len(msgs))
        muid_map = {}
        for msg in msgs:
            msgno, muid = msg.split(' ')
            muid_map[muid] = msgno
            if muid in downloaded:
                continue
            paccount.last_msg = now
            paccount.log.append(Pop3Log(muid=muid, date=now, paccount=paccount))
            self.log.debug('Downloading message ' + muid)
            response, lines, octets = server.retr(msgno)
            message = email.message_from_string('\r\n'.join(lines))
            msg = self.inject_message(paccount.account, message)
            if paccount.archive:
                msg.conv.remove_tag('Inbox')
            if paccount.tag:
                msg.conv.add_tag(paccount.tag)
        # Prune old messages from server
        for item in paccount.log:
            if not item.muid in muid_map:
                objectstore.delete(item)
            elif paccount.keep_time < 0:
                continue
            elif item.date <= now - timedelta(seconds=paccount.keep_time):
                self.log.debug('Deleting old POP3 message: ' + item.muid)
                server.dele(muid_map[item.muid])
                objectstore.delete(item)
        server.quit()
    

class FetchCommand(AdminCommand):
    cmd_name = 'fetch'
    short_help = "Fetch messages from POP3 servers"
    usage = '%prog fetch [options]'
    run_in_transaction = False

    def setup_optparser(self):
        self.parser.add_option('-n', '--interval', type='int', 
                               dest='interval', default=300,
                               help='Delay in seconds between message checks '
                               '(default: 300)')
        self.parser.add_option('-o', '--once', action='store_true', 
                               dest='once', default=False,
                               help="Just check one, don't loop forever")
 
    def finalize_options(self):
        if len(self.args) != 0:
            self.parser.error('Invalid number of arguments')
    
    def run(self):
        fetcher = POP3Fetcher(self.instance, once=self.options.once,
                              interval=self.options.interval)
        fetcher.run()
