# -*- coding: utf-8 -*-
#
# Copyright (C) 2006-2007 Edgewall Software
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://posterity.edgewall.org/wiki/License.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://posterity.edgewall.org/log/.

import re
import sys
import time
import mailbox
import email
from datetime import datetime
from optparse import OptionParser

from sqlalchemy import *
from elixir import metadata, objectstore

from posterity.admin import AdminCommand
from posterity.core import PosterityApplication
from posterity.model import Account, Msg
from posterity.util.mail import parse_rfc2822_date


class Injector(object):
    
    def __init__(self, instance):
        self.instance = instance
        self.log = self.instance.log
    
    def inject_mbox(self, account, fp, sort=False):
        def msgfactory(fp):
            try:
                return email.message_from_file(fp)
            except email.Errors.MessageParseError:
                # Don't return None since that will
                # stop the mailbox iterator
                return ''
        mbox = mailbox.UnixMailbox(fp, msgfactory)
        msgs = []
        # The messages needs to be in the correct order for the conversation
        # detection algorithms to work best
        if sort:
            for msg in mbox:
                msgs.append(msg)
        
            msgs.sort(lambda a, b: cmp(a['date'] and parse_rfc2822_date(a['date']) or datetime.utcnow(),
                                       b['date'] and parse_rfc2822_date(b['date']) or datetime.utcnow()))
        else:
            msgs = mbox

        for i, msg in enumerate(msgs):
            self.inject_message(account, msg)
            objectstore.flush()
            print i

    def inject_message(self, account, message):
        msg = Msg.parse_message(account, message)
        self.apply_filters(account, msg, message)
        return msg

    def apply_filters(self, account, msg, message):
        conv = msg.conv
        conv.add_tag('All Mail')
        conv.add_tag('Inbox')
        if message['X-DSPAM-Result'] == 'Spam':
            conv.add_tag('Spam')
            msg.is_deleted = True
        for filter in self.instance.message_filters:
            filter.incoming_message(account, msg)
        conv._update_date_cache() 


class InjectCommand(AdminCommand):
    cmd_name = 'inject'
    short_help = "Inject email messages into an ACCOUNT"
    usage = '%prog inject [options] account'

    def setup_optparser(self):
        self.parser.add_option("-f", "--file", dest="file",
                               help="read mail data from FILE instead of STDIN", 
                               metavar="FILE")
        self.parser.add_option("-m", "--mbox", dest="mbox", action="store_true",
                               help="inject mbox mailbox", default=False)
        self.parser.add_option("-q", "--quiet", action="store_false", 
                               dest="verbose", default=True,
                               help="don't print status messages to stdout")
 
    def finalize_options(self):
        if len(self.args) != 1:
            self.parser.error('Invalid number of arguments')
        self.account = self.args[0]
    
    def run(self):
        account = Account.get_by(username=self.account)
        if not account:
            self.parser.error('Invalid/missing username: ' + self.account)
        injector = Injector(self.instance)
        if self.options.file:
            fp = open(self.options.file, 'rb')
        else:
            fp = sys.stdin
        if self.options.mbox:
            injector.inject_mbox(account, fp, sort=True)
        else:
            message = email.message_from_file(fp)
            injector.inject_message(account, message)
